/*++

Copyright (c) 2010-11 OKI SEMICONDUCTOR CO.,LTD.  All Rights Reserved

Module Name:

	can_sample.c

Abstract:

	Sample program for CAN on Intel(R) Platform Controller Hub EG20T.
	Helper program (Loopback) for CAN device driver sample program.

Environment:

	Win32 console multi-threaded application

	Specify this exe before can_test2 TP on another machine.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.

--*/

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strsafe.h>
#include <initguid.h>
#include <setupapi.h>

#include "ioh_can_common.h"
#include "ioh_can_ioctls.h"

/*Define function input and output parameters*/
BOOL RunCANSample(__in LPCTSTR DriverName, __in int argc, __in_ecount(argc) char *argv[]);
BOOL RunCANLoopback(__in LPCTSTR DriverName);
BOOL GetDevicePath(__in LPGUID pGuid, __out PCHAR pDeviceName);
BOOL bFilterEnable = TRUE;

void print_usage(void)
{
}

int __cdecl
main(
	__in int argc,
	__in_ecount(argc) char *argv[]
	)
{
	CHAR DevicePath[MAX_PATH];
	/*Loop*/
	if((argc == 2) && !strcmp(argv[1], "-loop"))
	{
		GetDevicePath((LPGUID)&GUID_DEVINTERFACE_IOHCAN, DevicePath);
		RunCANLoopback(DevicePath);
	}

	/*Send*/
	else if((argc >= 2) && !strcmp(argv[1], "-send"))
	{
		GetDevicePath((LPGUID)&GUID_DEVINTERFACE_IOHCAN, DevicePath);
		RunCANSample(DevicePath, argc-1, argv+2);
	}
	else
	{
		printf("\n");
		printf("Usage: %s [-send [value] | -loop]\n", argv[0]);
		printf(" -send [value]  Transmit data, and receive. \n");
		printf("                Specify the initial value of the transmit data (value=0-255).\n");
		printf("                The default value is 0.\n");
		printf(" -loop          Operate as loop back program. \n");
		printf("\n");
	}
	return 0;
}

/**************************************************************************************************
Function: Obtain CAN device path

-Allocate CAN device length
-Retrieve CAN device path
**************************************************************************************************/
BOOL
GetDevicePath(
	LPGUID pGuid,PCHAR pDeviceName
	)
{
	HDEVINFO	hardwareDeviceInfo = NULL;
	SP_DEVICE_INTERFACE_DATA			deviceInterfaceData ={0};
	PSP_DEVICE_INTERFACE_DETAIL_DATA	deviceInterfaceDetailData = NULL;
	ULONG								predictedLength = 0;
	ULONG								requiredLength = 0,bytes=0;
	int i =0;
	BOOL bRet = FALSE;

	
	/*Return a handle to a device information set that contains requested device information elements for a computer */
	hardwareDeviceInfo = SetupDiGetClassDevs(
						  pGuid,
						  NULL,
						  NULL,
						 (DIGCF_PRESENT | // Only Devices present
						 DIGCF_DEVICEINTERFACE)); // Function class devices.
	
	/*Error handling - Invalid device handle - fail to handle a device information */
	if( INVALID_HANDLE_VALUE == hardwareDeviceInfo )
	{
		printf("SetupDiGetClassDevs failed: %x\n", (UINT)GetLastError());
		return bRet;
	}

	deviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

	
	do {
		/*enumerates the device interface that are contained in a device information set*/
		if (SetupDiEnumDeviceInterfaces (hardwareDeviceInfo, /*A pointer to a device information set that contains the device interfaces for which to return information*/
								 0, /*A pointer that specifies a device information element in DeviceInfoSet - No care about specific PDOs*/
								 pGuid, /*A pointer to a GUID that specifies the device interface class */
								 i, /*A zero-based index into the list of interfaces in the device information set.*/
								 &deviceInterfaceData /*A pointer to a caller-allocated buffer that contains, on successful return, a completed SP_DEVICE_INTERFACE_DATA structure */))
		{
			//
			// Allocate a function class device data structure to
			// receive the information about this particular device.
			//

			//
			// First find out required length of the buffer
			//

			if(!SetupDiGetDeviceInterfaceDetail (
					hardwareDeviceInfo,
					&deviceInterfaceData,
					NULL, // probing so no output buffer yet
					0, // probing so output buffer length of zero
					&requiredLength,
					NULL)){ 
							// not interested in the specific dev-node
							if(ERROR_INSUFFICIENT_BUFFER != GetLastError())
							{
								printf("SetupDiGetDeviceInterfaceDetail failed %d\n", (INT)GetLastError());
								SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
								break;
							}
			}

			/*Obtain required length*/
			predictedLength = requiredLength; 

			/*Allocate length in device information set (SP_DEVICE_INTERFACE_DATA) */
			deviceInterfaceDetailData = malloc (predictedLength); 

			if(deviceInterfaceDetailData)
			{
				/*Set size of device information set (SP_DEVICE_INTERFACE_DATA) in structure contains 
				the path for a device interface (deviceInterfaceDetailData)*/
				deviceInterfaceDetailData->cbSize =
								sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);
			} 
			else
			{
				/*Error handling - unable to allocate memory size in device inteface*/
				printf("Couldn't allocate %d bytes for device interface details.\n", (INT)predictedLength);
				SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
				break;
			}


			/*Error handling - Fail to returns details about a device interface*/
			if (! SetupDiGetDeviceInterfaceDetail (
					   hardwareDeviceInfo,
					   &deviceInterfaceData,
					   deviceInterfaceDetailData,
					   predictedLength,
					   &requiredLength,
					   NULL)){
								printf("Error in SetupDiGetDeviceInterfaceDetail\n");
								SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
								free (deviceInterfaceDetailData);
								break;
			}

			printf("%d) %s\n", ++i, deviceInterfaceDetailData->DevicePath);

			/*Copy device path to a variable called pDeviceName*/
			strcpy_s(pDeviceName,MAX_PATH,deviceInterfaceDetailData->DevicePath);
			bRet = TRUE;
			break;
		}

		/*Error handling - unable to enumerates device interface*/
		else if (ERROR_NO_MORE_ITEMS != GetLastError())
		{
			free (deviceInterfaceDetailData);
			deviceInterfaceDetailData = NULL;
			break;
		}
	} while (FALSE);

	/*Deletes a device information set and frees all associated memory*/
	SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);

	/*Error handling - device unavailable*/
	if(!deviceInterfaceDetailData)
	{
		printf("No device interface present\n");
		bRet = FALSE;
	}

	return bRet;
}

/*Function of printing message*/
void print_msg(const ioh_can_msg_t *msg)
{
	int i;
	printf("      msg.ide = %02d\n", msg->ide);
	printf("      msg.id = %04X\n", msg->id);
	printf("      msg.dlc = %d\n", msg->dlc);
	for (i=0; i<msg->dlc; i++) {
		printf("      msg.data[%d] = %02Xh\n", (INT)i, (UINT)msg->data[i]);
	}
	printf("      msg.rtr = %d\n", msg->rtr);
}


// Run CAN Device Driver Check
/**************************************************************************************************
Actor : Sender
Function: Test CAN controller interface - Sending message data via CAN controller interface

-Open an I/O device
-Set memory buffer value
-Assign user input as Data and other message information (IDE,ID,DLC,RTR) into memory buffer
-The user input value will be increment by 1 each time and insert into 8 different memory buffers (DLC=8)
-Write CAN message to the receiver side via CAN controller interface
-Retrieve the acknowledgement from receiver side by displaying out the CAN message that receive by receiver

Expected Result:

User input will be converted into hexadecimal and inserted into memory buffer. Each consecutive byte is adding by 1.
For example,
If user input is 123 (CAN_SAMPLE.exe -send 123), it will insert 7Bh,7Ch,7Dh,7Eh,7Fh,80h,81h,82h into the memory buffer.
While, if the user input is none (CAN_SAMPLE.exe -send), it will insert 00h,01h,02h,03h,04h,05h,06h,07h into the memory buffer - function will assume user input is 0.

**************************************************************************************************/
BOOL
RunCANSample(
	__in LPCTSTR DriverName,
	__in int argc,
	__in_ecount(argc) char *argv[]
	)
{
	UCHAR msgbuf[1];
	HANDLE hDevice;
	BOOL bRc = FALSE;
	DWORD errNum = 0;
	UCHAR value = 0;
	DWORD i,j;
	
	/*Declare 4 different message ID*/
	DWORD id[] = {0x00a, 0x0aa};
	
	ULONG ulBlockGet=0;
	unsigned int uiReceiveBuffNo=1;
	// DeviceIoControl parameters
	DWORD dwBytesReturned;

	// For CAN Device Driver API
	ioh_can_baud_t baud;
	ioh_can_msg_t msg = {0};
	ioh_can_timing_t timing;
	ioh_can_error_t errorStat;

	//CAN Filter Configuration @
	//ioh_can_rx_filter_t filter;
	
	printf("\n=========================\n");
	printf("Sample program for CAN device driver (%s) \n", DriverName);
	printf("=========================\n");
	
	if (argc>0) {
		value = (UCHAR)atoi(argv[0]);
		printf("Specified value : %02X\n",value);
	}

	/*Open an I/O device*/
	printf("\n*** CreateFile(%s) \n", DriverName);
	hDevice = CreateFile(DriverName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0,NULL);
	if(hDevice != INVALID_HANDLE_VALUE) {
		printf("  CreateFile Successful %s (hDevice:%08X)\n", DriverName, (UINT)hDevice);
	}
	else{	/*Error handling in opening an I/O device (CreateFile)*/
			printf("Failed CreateFile %s (LastError:%d)\n", DriverName, (INT)GetLastError());
			goto endSample;
	}

//---------------------------------------------------------------		

	/*Send a control code directly to a specified device driver 
	IOCTL_CAN_STOP: Stop all operations and the device, this is to enable
	new/custom setting to be set in the device driver*/	
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_STOP)\n");
	bRc = DeviceIoControl(hDevice, IOCTL_CAN_STOP, NULL, 0, NULL, 0, &dwBytesReturned, NULL);
	if(bRc == TRUE){
		printf("  DeviceIoControl Successful (IOCTL_CAN_STOP)\n");
		dwBytesReturned = 0;
	}
	else if (bRc == FALSE) {
		/*Error handling in sending a control code to a specified device driver (DeviceIoControl)*/
			errNum = GetLastError(); //retrieve error code value
			printf("  Failed DeviceIoControl GetLastError:%d\n", (INT)errNum);
			goto failedSample;
	}	
	
	/*Send a control code directly to a specified device driver - Set pre-defined recommended timing (baud rate)*/
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_SIMPLE)\n");
	baud = IOH_CAN_BAUD_10;
	bRc = DeviceIoControl(hDevice,IOCTL_CAN_SIMPLE,&baud,sizeof(baud),NULL,0,&dwBytesReturned,NULL);
	if(bRc == TRUE) {
		printf("  DeviceIoControl Successful (IOCTL_CAN_SIMPLE)\n");
		printf("      dwBytesReturned = %ld\n", dwBytesReturned);
		dwBytesReturned = 0;	
	}else{goto failedSample;}

	/*Send a control code directly to a specified device driver - Set the device to run (After configured the setting)*/
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_RUN)\n");
	bRc = DeviceIoControl(hDevice, IOCTL_CAN_RUN, NULL, 0, NULL, 0, &dwBytesReturned, NULL);
	if(bRc == TRUE){
	printf("  DeviceIoControl Successful (IOCTL_CAN_RUN)\n");
	dwBytesReturned = 0;
	}else{goto failedSample;}
	
	
	/*Send a control code directly to a specified device driver.  
	Able to select between block or non block mode whereby in block mode write operation will 
	not return until the message has been transmitted. 
	In non block mode, the write operation will return immeadiately, 
	which means it does not wait until the message is fully transmitted
	Note: This function need to be called before IOCTL_CAN_WRITE function*/
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_BLOCK)\n");
	bRc= DeviceIoControl(hDevice,IOCTL_CAN_BLOCK,NULL,0,NULL,0,&dwBytesReturned,NULL);
	if(bRc == TRUE){
		printf(" DeviceIoControl Successful (IOCTL_CAN_BLOCK)\n");
		printf("      dwBytesReturned = %ld\n", dwBytesReturned);
		dwBytesReturned = 0;
	}else{goto failedSample;}	
		
		
	/*Send a control code directly to a specified device driver - Get Blocking state of device 
	ulBlockGet = 0 - Non Block Mode
	ulBlockGet = 1 - Block Mode*/
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_BLOCK_GET)\n");
	bRc= DeviceIoControl(hDevice,IOCTL_CAN_BLOCK_GET,NULL,0,&ulBlockGet,sizeof(ulBlockGet),&dwBytesReturned,NULL);
	if(bRc==TRUE){
		printf(" DeviceIoControl Successful (IOCTL_CAN_BLOCK_GET)\n");
		printf("      dwBytesReturned = %ld\n", dwBytesReturned);
		printf("Block Mode : %ld\n",ulBlockGet);
		dwBytesReturned = 0;
	}else{goto failedSample;}	
		
	
	/*Send a control code directly to a specified device driver - Get current bit timing values */
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_TIMING_GET)\n");
	bRc= DeviceIoControl(hDevice,IOCTL_CAN_TIMING_GET,NULL,0,&timing,sizeof(ioh_can_timing_t),&dwBytesReturned,NULL);
	if(bRc==TRUE){
		printf(" DeviceIoControl Successful (IOCTL_CAN_TIMING_GET)\n");
		printf("      dwBytesReturned = %ld\n", dwBytesReturned);
		printf("Timing : %d\n",timing.cfg_bitrate);
		dwBytesReturned = 0;
	
		
		/*Assign message data structure with all the neccessary information*/
		for(j = 0; j < _countof(id); j++){
			memset(&msg, 0, sizeof(msg));
			msg.ide = 0;
			msg.id = id[j];
			msg.dlc = 8;
			for (i=0; i<msg.dlc; i++) {
				msg.data[i] = (unsigned char)(i + (j * 8) + value);
			}
			msg.rtr = 0;
			

			printf("\n*** CAN WRITING(code:IOCTL_CAN_WRITE)\n");
			/*Call printing message function - Display out message data that ready to be send*/
			print_msg(&msg);
			/*Send a control code directly to a specified device driver - Transmit CAN message to the receiver side via CAN controller interface*/
			bRc = DeviceIoControl(hDevice,IOCTL_CAN_WRITE,&msg,sizeof(msg),NULL,0,&dwBytesReturned,NULL);
			printf("\n*** CAN WRITING(code:IOCTL_CAN_WRITE IDE:0, ID:7FFh, DLC:8, <%02Xh> ... <%02Xh>, RTR:0)\n", (unsigned int)value, (unsigned int)(value+7));
			if(bRc == TRUE) {
				printf(" DeviceIoControl Successful (IOCTL_CAN_WRITE)\n");
				printf("      dwBytesReturned = %ld\n", dwBytesReturned);
				Sleep(100); //Temporary on-hold the execution of the current function until the time-out interval elapses (milliseconds).
				}
		}
		dwBytesReturned = 0;
			
	}else{goto failedSample;}			
			
	printf("\n*** CAN READING(code:IOCTL_CAN_READ)\n");
	for(j=0;j<_countof(id);j++){
		printf("\n*** Waiting for CAN data receiving(code:IOCTL_CAN_READ)\n");
		memset(&msg, 0, sizeof(msg));
		/*Send a control code directly to a specified device driver - Retrieve and display out the message data received by the receiver*/
		bRc = DeviceIoControl(hDevice,IOCTL_CAN_READ,NULL,0,&msg,sizeof(msg),&dwBytesReturned,NULL);
		if(bRc == TRUE) {
			printf("  DeviceIoControl Successful (IOCTL_CAN_READ)\n");
			printf("      dwBytesReturned = %ld\n", dwBytesReturned);
		}else{continue;}	
		if ( (dwBytesReturned==sizeof(msg)) && (msg.dlc>0) ) {
			/*Call printing message function - Display out the message data receiver by receive*/
			print_msg(&msg);	
		}
	}	
//-------------------------------------------------------------
failedSample:
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_ERROR_STATS_GET)\n");
	/*Send a control code directly to a specified device driver - Get error status of CAN device*/
	dwBytesReturned = 0;
	bRc = DeviceIoControl(hDevice,IOCTL_CAN_ERROR_STATS_GET,NULL,0,&errorStat,sizeof(ioh_can_error_t),&dwBytesReturned,NULL);
	if(bRc == TRUE) {
		printf("  DeviceIoControl Successful (IOCTL_CAN_ERROR_STATS_GET)\n");
		printf("      dwBytesReturned = %ld\n", dwBytesReturned);
		printf(" Receive error count (>=96) : %d\n",errorStat.rxgte96);
		printf(" Transmit error count (>=96): %d\n",errorStat.txgte96);
		printf(" Error state : %d\n",errorStat.error_stat);
		printf(" Receive error count : %d\n",errorStat.rx_err_cnt);
		printf(" Transmit error count : %d\n",errorStat.tx_err_cnt);
	}

//----DEVICE HANDLE START---
	/*Close device handle*/
	printf("\n*** CloseHandle(hDevice:%08X)\n", (UINT)hDevice);
	if(CloseHandle(hDevice)) {
		printf("  CloseHandle Successful\n");
		bRc =  TRUE;
	}
	/*Error handling in close device handle function*/
	else {
		errNum = GetLastError();
		printf("  Failed CloseHandle GetLastError:%d\n", (INT)errNum);
	}
//----DEVICE HANDLE END-----
 endSample:	
	return bRc;
}


// Run CAN Device Driver Check
/**************************************************************************************************
Actor : Receiver
Function: Test CAN controller interface - receiving message data via CAN controller interface

-Open a I/O device
-Set memory buffer size
-Retrieve and display out all the message receive from the sender
-Return byte to the sender as a form of acknowledgement

Expected Result (CAN_SAMPLE.exe -loop)

For example,
If receiver receive the value of 123, it will display out 7Bh,7Ch,7Dh,7Eh,7Fh,80h,81h,82h in array form.
If receiver receive the value of 0, it will display out 00h,01h,02h,03h,04h,05h,06h,07h in array form.
Each consecutive byte is adding by 1
**************************************************************************************************/
BOOL
RunCANLoopback(
	__in LPCTSTR DriverName
	)
{
	UCHAR msgbuf[1];
	HANDLE hDevice;
	BOOL bRc = FALSE;
	DWORD errNum = 0;
	// UCHAR value = 0;
	
	ioh_can_error_t errorStat;

	unsigned int uiReceiveBuffNo=1;
	
	// DeviceIoControl parameters
	DWORD dwBytesReturned;

	// For CAN Device Driver API
	ioh_can_baud_t baud;
	ioh_can_msg_t msg = {0};

	//CAN Filter Configuration @
	ioh_can_rx_filter_t filter;

	
	printf("\n=========================\n");
	printf("Helper program (loop back) for CAN device driver (%s) \n", DriverName);
	printf("=========================\n");
	
	/*Open an I/O device*/
	printf("\n*** CreateFile(%s) \n", DriverName);
	hDevice = CreateFile(DriverName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0,NULL);
	if(hDevice != INVALID_HANDLE_VALUE) {
		printf("  CreateFile Successful %s (hDevice:%08X)\n", DriverName, (UINT)hDevice);
		baud = IOH_CAN_BAUD_10; // 10Kbps
		dwBytesReturned = 0;
	}
	else {
		/*Error handling in opening an I/O device (CreateFile)*/
		printf("Failed CreateFile %s (LastError:%d)\n", DriverName, (INT)GetLastError());
		goto endLoopback;
	}	


	/*Send a control code directly to a specified device driver - Stop all operations and the device
	Configure CAN driver setting*/	
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_STOP)\n");
	bRc = DeviceIoControl(hDevice, IOCTL_CAN_STOP, NULL, 0, NULL, 0, &dwBytesReturned, NULL);
	if(bRc==TRUE){
	printf("  DeviceIoControl Successful (IOCTL_CAN_STOP)\n");
	printf("      dwBytesReturned = %ld\n", dwBytesReturned);
	}else{goto failedLoopback;}	
			
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_SIMPLE)\n");
	/*Send a control code directly to a specified device driver - Set pre-defined recommended timing (baud rate)*/
	bRc = DeviceIoControl(hDevice,IOCTL_CAN_SIMPLE,&baud,sizeof(baud),NULL,0,&dwBytesReturned,NULL);
	if(bRc == TRUE) {
		printf(" DeviceIoControl Successful (IOCTL_CAN_SIMPLE)\n");
		printf("      dwBytesReturned = %ld\n", dwBytesReturned);
		memset(&msg, 0, sizeof(msg));
		dwBytesReturned = 0;
	}else{goto failedLoopback;}			
			
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_RX_ENABLE_CLEAR)\n", DriverName);
	/*Looping to clear up all the 16 receive buffers in CAN controller*/
	for(uiReceiveBuffNo = 1; uiReceiveBuffNo <= 16; uiReceiveBuffNo++){
		/*Send a control code directly to a specified device driver - Clear (disable) a receive buffer*/
		bRc = DeviceIoControl(hDevice, IOCTL_CAN_RX_ENABLE_CLEAR, &uiReceiveBuffNo, sizeof(uiReceiveBuffNo), NULL, 0, &dwBytesReturned, NULL);
		if(bRc==TRUE){
			printf(" DeviceIOControl Successful (IOCTL_CAN_RX_ENABLE_CLEAR)\n");
		}else{printf("Error: Buffer failed to clear");}
	}
	dwBytesReturned = 0;
	
	
	uiReceiveBuffNo = 1;	// only Receive Buffer 1 is used
	if(!bFilterEnable){goto failedLoopback;}
	printf("\nSet Filtering Section\n");
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_FILTER)\n", DriverName);
	/*Filter Configuration*/
	filter.num=uiReceiveBuffNo;
	filter.umask=1;			// use mask
	filter.amr.id=0x00f;	// 4 LSB of ID is used for receive filter (acceptance filtering of received messages)
	filter.amr.id_ext=0;	// standard CAN frame (11 bit) ID
	filter.amr.rtr=0;		// 0 remote transmission request
	filter.aidr.id=0x00a;	// 4 LSB of ID is 1010
	filter.aidr.id_ext=0;	// standard CAN frame (11 bit) ID
	filter.aidr.rtr=0;		// 0 remote transmission request
	
	
	
	/* Setting the CAN message filter based on the above filter structure for a particular receive buffer*/
	/* The DeviceIoControl send a control code (IOCTL_CAN_FILTER with input parameter (&filter))
	   directly to a CAN driver*/
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_FILTER)\n");
	bRc = DeviceIoControl(hDevice, IOCTL_CAN_FILTER, &filter, /*Input parameter*/
							sizeof(ioh_can_rx_filter_t),
							NULL, 0, &dwBytesReturned, NULL);
	if(bRc==TRUE){
		printf(" DeviceIoControl Successful (IOCTL_CAN_FILTER)\n");
		dwBytesReturned = 0; /*This value need to be reset*/
	}else{goto failedLoopback;}

	
	/*To enable back the CAN RX Filter, to ensure only relevant filtered messages are stored in the buffer*/
	/*DeviceIoControl will send a control code (IOCTL_CAN_RX_ENABLE_SET with input parameter (uiReceiveBuffNo))
	  directly to a CAN driver */
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_RX_ENABLE_SET)\n");
	bRc = DeviceIoControl(hDevice, IOCTL_CAN_RX_ENABLE_SET, &uiReceiveBuffNo, 
		sizeof(uiReceiveBuffNo), NULL, 0, &dwBytesReturned, NULL);
	if(bRc==TRUE){
		printf(" DeviceIoControl Successful (IOCTL_CAN_RX_ENABLE_SET)\n");
		dwBytesReturned = 0;
	}else{goto failedLoopback;}

	
	
	/*Send a control code directly to a specified device driver - Set the device to run (After configured the setting)*/
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_RUN)\n");
	bRc = DeviceIoControl(hDevice, IOCTL_CAN_RUN, NULL, 0, NULL, 0, &dwBytesReturned, NULL);
	if(bRc==TRUE){
		printf(" DeviceIoControl Successful (IOCTL_CAN_RUN)\n");
		printf("      dwBytesReturned = %ld\n", dwBytesReturned);
	}else{goto failedLoopback;}

	
	/*Send a control code directly to a specified device driver - Able to select between block or non block mode 
	whereby in block mode read operation will wait until a bufffer is received. 
	In non block mode, the read operation will return immediately, if a message is waiting, read operation will return with that message
	Note: This function need to be called before IOCTL_CAN_READ function*/
	printf("\n*** DeviceIoControl(code:IOCTL_CAN_BLOCK)\n");
	bRc= DeviceIoControl(hDevice,IOCTL_CAN_BLOCK,NULL,0,NULL,0,&dwBytesReturned,NULL);
	if(bRc == TRUE){
		printf(" DeviceIoControl Successful (IOCTL_CAN_BLOCK)\n");
		printf("      dwBytesReturned = %ld\n", dwBytesReturned);
	}else {goto failedLoopback;}
	
	while(bRc == TRUE) {
		printf("\n*** Waiting for CAN data receiving(code:IOCTL_CAN_READ)\n");
		dwBytesReturned = 0;
		
		
		/*Send a control code directly to a specified device driver - Retrieve and display out message data received*/
		printf("\n*** DeviceIoControl(code:IOCTL_CAN_READ)\n");
		bRc = DeviceIoControl(hDevice,IOCTL_CAN_READ,NULL,0,&msg,sizeof(msg),&dwBytesReturned,NULL);
		if(bRc == TRUE) {
			printf("  DeviceIoControl Successful (IOCTL_CAN_READ)\n");
			printf("      dwBytesReturned = %ld\n", dwBytesReturned);
		}else{continue;}
		if ( (dwBytesReturned==sizeof(msg)) && (msg.dlc>0) ) {
			print_msg(&msg);
			dwBytesReturned = 0;
		}else{continue;}

		
		/*Send a control code directly to a specified device driver - Send data received as acknowledgement back to sender*/
		printf("\n*** DeviceIoControl(code:IOCTL_CAN_WRITE)\n");					
		bRc = DeviceIoControl(hDevice,IOCTL_CAN_WRITE,&msg,sizeof(msg),NULL,0,&dwBytesReturned,NULL);
		if(bRc == TRUE) {
			printf("  DeviceIoControl Successful (IOCTL_CAN_WRITE)\n");
			printf("      dwBytesReturned = %ld\n", dwBytesReturned);			
			dwBytesReturned = 0;
		}else{continue;}

												
		/*Send a control code directly to a specified device driver - Get error status of CAN device
		Condition: Successfully read CAN message
		Any transmit error including ACK error occurred, transmit node immediately resend the same message until transmit succeed.
		Please refer to CAN Spec for error count value code.
		*/
		printf("\n*** DeviceIoControl(code:IOCTL_CAN_ERROR_STATS_GET)\n");
		bRc = DeviceIoControl(hDevice,IOCTL_CAN_ERROR_STATS_GET,NULL,0,&errorStat,sizeof(ioh_can_error_t),&dwBytesReturned,NULL);
		if(bRc == TRUE) {
			printf("  DeviceIoControl Successful (IOCTL_CAN_ERROR_STATS_GET)\n");
			printf("      dwBytesReturned = %ld\n", dwBytesReturned);
			printf(" Receive error count (>=96) : %d\n",errorStat.rxgte96);
			printf(" Transmit error count (>=96): %d\n",errorStat.txgte96);
			printf(" Error state : %d\n",errorStat.error_stat);
			printf(" Receive error count : %d\n",errorStat.rx_err_cnt);
			printf(" Transmit error count : %d\n",errorStat.tx_err_cnt);
		}
	}
		
	


failedLoopback:
	/*Error handling in sending a control code to a specified device driver (DeviceIoControl)*/
	if (bRc == FALSE) {
		errNum = GetLastError();
		printf("  Failed DeviceIoControl GetLastError:%d\n", (INT)errNum);
	}
			
	
	/*Close device handle*/
	printf("\n*** CloseHandle(hDevice:%08X)\n", (UINT)hDevice);
	if(CloseHandle(hDevice)) {
		printf("  CloseHandle Successful\n");
		bRc =  TRUE;
	}
	

	/*Error handling in close device handle function*/
	else {
		errNum = GetLastError();
		printf("  Failed CloseHandle GetLastError:%d\n", (INT)errNum);
	}
endLoopback:
	return bRc;
}
