/*++

Copyright (c) 2010-11 OKI SEMICONDUCTOR CO.,LTD.  All Rights Reserved

Module Name:

	usb_sample.c

Abstract:
	Sample program for USB Client on Intel(R) Platform Controller Hub EG20T.

Environment:

	Win32 console multi-threaded application

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.

--*/

#include <windows.h>
#include <stddef.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <winioctl.h>
#include <usb200.h>
#include "ioh_udc_ioctls.h"

// This defines the IOCTL constants.
#include <dontuse.h>
#include <setupapi.h>
#include <initguid.h>

#pragma warning( disable : 4995 )
#pragma warning( disable : 4242 )
#pragma warning( disable : 4715 )


#define BUFFER_SIZE (4096)	/* Max. 4096 Bytes */

#define USB_STRING_SIZE1 1
#define USB_STRING_SIZE2 7
#define USB_STRING_SIZE3 18

#pragma pack(1)
typedef struct _USB_DESCRIPTOR_SAMPLE {
	USB_DEVICE_DESCRIPTOR				usbDeviceDesc;
	USB_DEVICE_QUALIFIER_DESCRIPTOR 	usbDeviceQualDesc;

	struct _HighSpeedConfig1 {
		USB_CONFIGURATION_DESCRIPTOR	usbConfigDesc;
		USB_INTERFACE_DESCRIPTOR		usbInterfaceDesc;
		USB_ENDPOINT_DESCRIPTOR 		usbEndPointDesc1;
		USB_ENDPOINT_DESCRIPTOR 		usbEndPointDesc2;
	}	HSConfig;

	struct _FullSpeedConfigg1 {
		USB_CONFIGURATION_DESCRIPTOR	usbOSpConfigDesc;
		USB_INTERFACE_DESCRIPTOR		usbInterfaceDesc;
		USB_ENDPOINT_DESCRIPTOR 		usbEndPointDesc1;
		USB_ENDPOINT_DESCRIPTOR 		usbEndPointDesc2;
	}	FSConfig;

	struct _UsbStringDescriptor1 {
		BYTE bLength;
		BYTE bDescriptorType;
		WORD String[USB_STRING_SIZE1];
	}	usbStringDesc1;

	struct _UsbStringDescriptor2 {
		BYTE bLength;
		BYTE bDescriptorType;
		WORD String[USB_STRING_SIZE2];
	}	usbStringDesc2;

	struct _UsbStringDescriptor3 {
		BYTE bLength;
		BYTE bDescriptorType;
		WORD String[USB_STRING_SIZE3];
	}	usbStringDesc3;

} USB_DESCRIPTOR_SAMPLE, *PUSB_DESCRIPTOR_SAMPLE;
#pragma pack()


BOOL GetDevicePath(LPGUID pGuid,PCHAR pDeviceName)
{
	HDEVINFO	hardwareDeviceInfo = NULL;
	SP_DEVICE_INTERFACE_DATA			deviceInterfaceData ={0};
	PSP_DEVICE_INTERFACE_DETAIL_DATA	deviceInterfaceDetailData = NULL;
	ULONG								predictedLength = 0;
	ULONG								requiredLength = 0,bytes=0;
	int i =0;
	BOOL bRet = FALSE;

	//
	// Open a handle to the device interface information set of all
	// present	system class interfaces.
	//

	hardwareDeviceInfo = SetupDiGetClassDevs(
						  pGuid,
						  NULL,
						  NULL,
						 (DIGCF_PRESENT | // Only Devices present
						 DIGCF_DEVICEINTERFACE)); // Function class devices.

	if( INVALID_HANDLE_VALUE == hardwareDeviceInfo )
	{
		printf("SetupDiGetClassDevs failed: %d\n", (INT)GetLastError());
		return bRet;
	}

	deviceInterfaceData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

	//
	// Enumerate devices of USB Device class
	//

	do {

		if (SetupDiEnumDeviceInterfaces (hardwareDeviceInfo,
								 0, // No care about specific PDOs
								 pGuid,
								 i, //
								 &deviceInterfaceData))
			{

			//
			// Allocate a function class device data structure to
			// receive the information about this particular device.
			//

			//
			// First find out required length of the buffer
			//

			if(!SetupDiGetDeviceInterfaceDetail (
					hardwareDeviceInfo,
					&deviceInterfaceData,
					NULL, // probing so no output buffer yet
					0, // probing so output buffer length of zero
					&requiredLength,
					NULL))
			{ // not interested in the specific dev-node
				if(ERROR_INSUFFICIENT_BUFFER != GetLastError())
				{
					printf("SetupDiGetDeviceInterfaceDetail failed %d\n", (INT)GetLastError());
					SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
					break;
				}
			}

			predictedLength = requiredLength;

			deviceInterfaceDetailData = malloc (predictedLength);

			if(deviceInterfaceDetailData)
			{
				deviceInterfaceDetailData->cbSize =
								sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);
			} else
			{
				printf("Couldn't allocate %d bytes for device interface details.\n", (INT)predictedLength);
				SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
				break;
			}

			if (! SetupDiGetDeviceInterfaceDetail (
					   hardwareDeviceInfo,
					   &deviceInterfaceData,
					   deviceInterfaceDetailData,
					   predictedLength,
					   &requiredLength,
					   NULL))
			{
				printf("Error in SetupDiGetDeviceInterfaceDetail\n");
				SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);
				free (deviceInterfaceDetailData);
				break;
			}

			printf("%d) %s\n", ++i,
					deviceInterfaceDetailData->DevicePath);

			strcpy_s(pDeviceName, MAX_PATH, deviceInterfaceDetailData->DevicePath);
			bRet = TRUE;
			break;
		}
		else if (ERROR_NO_MORE_ITEMS != GetLastError())
		{
			free (deviceInterfaceDetailData);
			deviceInterfaceDetailData = NULL;
			break;
		}
	} while (FALSE);

	SetupDiDestroyDeviceInfoList (hardwareDeviceInfo);

	if(!deviceInterfaceDetailData)
	{
		printf("No device interfaces present\n");
		bRet = FALSE;
	}

	return bRet;
}


void usbDescriptorFill(PUSB_DESCRIPTOR_SAMPLE pUSBDesc)
{
	/* Filling Device Descriptor */
	pUSBDesc->usbDeviceDesc.bLength 						= sizeof(USB_DEVICE_DESCRIPTOR);
	pUSBDesc->usbDeviceDesc.bDescriptorType 				= USB_DEVICE_DESCRIPTOR_TYPE;
	pUSBDesc->usbDeviceDesc.bcdUSB							= 0x0200;
	pUSBDesc->usbDeviceDesc.bDeviceClass					= 0;
	pUSBDesc->usbDeviceDesc.bDeviceSubClass 				= 0;
	pUSBDesc->usbDeviceDesc.bDeviceProtocol 				= 0;
	pUSBDesc->usbDeviceDesc.bMaxPacketSize0 				= 0x40;
	pUSBDesc->usbDeviceDesc.idVendor						= 0x045e;
	pUSBDesc->usbDeviceDesc.idProduct						= 0x930a;
	pUSBDesc->usbDeviceDesc.bcdDevice						= 0x001;
	pUSBDesc->usbDeviceDesc.iManufacturer					= 1;
	pUSBDesc->usbDeviceDesc.iProduct						= 2;
	pUSBDesc->usbDeviceDesc.iSerialNumber					= 0;
	pUSBDesc->usbDeviceDesc.bNumConfigurations				= 1;

	/* Filling Device QualifierDescriptor */
	pUSBDesc->usbDeviceQualDesc.bLength 					= sizeof(USB_DEVICE_QUALIFIER_DESCRIPTOR);
	pUSBDesc->usbDeviceQualDesc.bDescriptorType 			= USB_DEVICE_QUALIFIER_DESCRIPTOR_TYPE;
	pUSBDesc->usbDeviceQualDesc.bcdUSB						= 0x0200;
	pUSBDesc->usbDeviceQualDesc.bDeviceClass				= 0;
	pUSBDesc->usbDeviceQualDesc.bDeviceSubClass 			= 0;
	pUSBDesc->usbDeviceQualDesc.bDeviceProtocol 			= 0;
	pUSBDesc->usbDeviceQualDesc.bMaxPacketSize0 			= 0x40;
	pUSBDesc->usbDeviceQualDesc.bReserved					= 0;
	pUSBDesc->usbDeviceQualDesc.bNumConfigurations			= 1;

	/* Filling Configuration  Descriptor*/
	pUSBDesc->HSConfig.usbConfigDesc.bLength				= sizeof(USB_CONFIGURATION_DESCRIPTOR);
	pUSBDesc->HSConfig.usbConfigDesc.bDescriptorType		= USB_CONFIGURATION_DESCRIPTOR_TYPE;
	pUSBDesc->HSConfig.usbConfigDesc.wTotalLength			= sizeof(USB_CONFIGURATION_DESCRIPTOR)
															+ sizeof(USB_INTERFACE_DESCRIPTOR)
															+ sizeof(USB_ENDPOINT_DESCRIPTOR) * 2;
	pUSBDesc->HSConfig.usbConfigDesc.bNumInterfaces 		= 1;
	pUSBDesc->HSConfig.usbConfigDesc.bConfigurationValue	= 1;
	pUSBDesc->HSConfig.usbConfigDesc.iConfiguration 		= 0;
	pUSBDesc->HSConfig.usbConfigDesc.bmAttributes			= 0xc0;
	pUSBDesc->HSConfig.usbConfigDesc.MaxPower				= 1;

	/* Filling Interface  Descriptor*/
	pUSBDesc->HSConfig.usbInterfaceDesc.bLength 			= sizeof(USB_INTERFACE_DESCRIPTOR);
	pUSBDesc->HSConfig.usbInterfaceDesc.bDescriptorType 	= USB_INTERFACE_DESCRIPTOR_TYPE;
	pUSBDesc->HSConfig.usbInterfaceDesc.bInterfaceNumber	= 0x0;
	pUSBDesc->HSConfig.usbInterfaceDesc.bAlternateSetting	= 0x0;
	pUSBDesc->HSConfig.usbInterfaceDesc.bNumEndpoints		= 0x2;
	pUSBDesc->HSConfig.usbInterfaceDesc.bInterfaceClass 	= 0xff;
	pUSBDesc->HSConfig.usbInterfaceDesc.bInterfaceSubClass	= 0xff;
	pUSBDesc->HSConfig.usbInterfaceDesc.bInterfaceProtocol	= 0xff;
	pUSBDesc->HSConfig.usbInterfaceDesc.iInterface			= 0x0;

	/* Filling Endpoint  Descriptor - 1*/
	pUSBDesc->HSConfig.usbEndPointDesc1.bLength 			= sizeof(USB_ENDPOINT_DESCRIPTOR);
	pUSBDesc->HSConfig.usbEndPointDesc1.bDescriptorType 	= USB_ENDPOINT_DESCRIPTOR_TYPE;
	pUSBDesc->HSConfig.usbEndPointDesc1.bEndpointAddress	= 0x81;		/* In */
	pUSBDesc->HSConfig.usbEndPointDesc1.bmAttributes		= 0x2;		/* Bulk */
	pUSBDesc->HSConfig.usbEndPointDesc1.wMaxPacketSize		= 0x200;	/* 512 bytes */
	pUSBDesc->HSConfig.usbEndPointDesc1.bInterval			= 0x1;

	/* Filling Endpoint  Descriptor - 2*/
	pUSBDesc->HSConfig.usbEndPointDesc2.bLength 			= sizeof(USB_ENDPOINT_DESCRIPTOR);
	pUSBDesc->HSConfig.usbEndPointDesc2.bDescriptorType 	= USB_ENDPOINT_DESCRIPTOR_TYPE;
	pUSBDesc->HSConfig.usbEndPointDesc2.bEndpointAddress	= 0x1;		/* Out */
	pUSBDesc->HSConfig.usbEndPointDesc2.bmAttributes		= 0x2;		/* Bulk */
	pUSBDesc->HSConfig.usbEndPointDesc2.wMaxPacketSize		= 0x200;	/* 512 bytes */
	pUSBDesc->HSConfig.usbEndPointDesc2.bInterval			= 0x0;

	/* Filling Configuration  Descriptor*/
	pUSBDesc->FSConfig.usbOSpConfigDesc.bLength 			= sizeof(USB_CONFIGURATION_DESCRIPTOR);
	pUSBDesc->FSConfig.usbOSpConfigDesc.bDescriptorType 	= USB_OTHER_SPEED_CONFIGURATION_DESCRIPTOR_TYPE;
	pUSBDesc->FSConfig.usbOSpConfigDesc.wTotalLength		= sizeof(USB_CONFIGURATION_DESCRIPTOR)
															+ sizeof(USB_INTERFACE_DESCRIPTOR)
															+ sizeof(USB_ENDPOINT_DESCRIPTOR) * 2;
	pUSBDesc->FSConfig.usbOSpConfigDesc.bNumInterfaces		= 1;
	pUSBDesc->FSConfig.usbOSpConfigDesc.bConfigurationValue = 1;
	pUSBDesc->FSConfig.usbOSpConfigDesc.iConfiguration		= 0;
	pUSBDesc->FSConfig.usbOSpConfigDesc.bmAttributes		= 0xc0;
	pUSBDesc->FSConfig.usbOSpConfigDesc.MaxPower			= 1;

	/* Filling Interface  Descriptor*/
	pUSBDesc->FSConfig.usbInterfaceDesc.bLength 			= sizeof(USB_INTERFACE_DESCRIPTOR);
	pUSBDesc->FSConfig.usbInterfaceDesc.bDescriptorType 	= USB_INTERFACE_DESCRIPTOR_TYPE;
	pUSBDesc->FSConfig.usbInterfaceDesc.bInterfaceNumber	= 0x0;
	pUSBDesc->FSConfig.usbInterfaceDesc.bAlternateSetting	= 0x0;
	pUSBDesc->FSConfig.usbInterfaceDesc.bNumEndpoints		= 0x2;
	pUSBDesc->FSConfig.usbInterfaceDesc.bInterfaceClass 	= 0xff;
	pUSBDesc->FSConfig.usbInterfaceDesc.bInterfaceSubClass	= 0xff;
	pUSBDesc->FSConfig.usbInterfaceDesc.bInterfaceProtocol	= 0xff;
	pUSBDesc->FSConfig.usbInterfaceDesc.iInterface			= 0x0;

	/* Filling Endpoint  Descriptor - 1*/
	pUSBDesc->FSConfig.usbEndPointDesc1.bLength 			= sizeof(USB_ENDPOINT_DESCRIPTOR);
	pUSBDesc->FSConfig.usbEndPointDesc1.bDescriptorType 	= USB_ENDPOINT_DESCRIPTOR_TYPE;
	pUSBDesc->FSConfig.usbEndPointDesc1.bEndpointAddress	= 0x81; 	/* IN */
	pUSBDesc->FSConfig.usbEndPointDesc1.bmAttributes		= 0x2;		/* Bulk */;
	pUSBDesc->FSConfig.usbEndPointDesc1.wMaxPacketSize		= 0x40;		/* 64 bytes */
	pUSBDesc->FSConfig.usbEndPointDesc1.bInterval			= 0x1;

	/* Filling Endpoint  Descriptor - 2*/
	pUSBDesc->FSConfig.usbEndPointDesc2.bLength 			= sizeof(USB_ENDPOINT_DESCRIPTOR);
	pUSBDesc->FSConfig.usbEndPointDesc2.bDescriptorType 	= USB_ENDPOINT_DESCRIPTOR_TYPE;
	pUSBDesc->FSConfig.usbEndPointDesc2.bEndpointAddress	= 0x1;		/* Out */
	pUSBDesc->FSConfig.usbEndPointDesc2.bmAttributes		= 0x2;		/* Bulk */;
	pUSBDesc->FSConfig.usbEndPointDesc2.wMaxPacketSize		= 0x40;		/* 64 bytes */
	pUSBDesc->FSConfig.usbEndPointDesc2.bInterval			= 0x0;

	/* Filling String  Descriptor */
	pUSBDesc->usbStringDesc1.bLength						= 2 + 2 * USB_STRING_SIZE1;
	pUSBDesc->usbStringDesc1.bDescriptorType				= USB_STRING_DESCRIPTOR_TYPE;
	pUSBDesc->usbStringDesc1.String[0]						= 0x0409;

	/* Filling String  Descriptor */
	pUSBDesc->usbStringDesc2.bLength						= 2 + 2 * USB_STRING_SIZE2;
	pUSBDesc->usbStringDesc2.bDescriptorType				= USB_STRING_DESCRIPTOR_TYPE;
	pUSBDesc->usbStringDesc2.String[0]						= 'V';
	pUSBDesc->usbStringDesc2.String[1]						= 'E';
	pUSBDesc->usbStringDesc2.String[2]						= 'N';
	pUSBDesc->usbStringDesc2.String[3]						= 'D';
	pUSBDesc->usbStringDesc2.String[4]						= 'O';
	pUSBDesc->usbStringDesc2.String[5]						= 'R';
	pUSBDesc->usbStringDesc2.String[6]						= '.';

	/* Filling String  Descriptor */
	pUSBDesc->usbStringDesc3.bLength						= 2 + 2 * USB_STRING_SIZE3;
	pUSBDesc->usbStringDesc3.bDescriptorType				= USB_STRING_DESCRIPTOR_TYPE;
	pUSBDesc->usbStringDesc3.String[0]						= 'U';
	pUSBDesc->usbStringDesc3.String[1]						= 'S';
	pUSBDesc->usbStringDesc3.String[2]						= 'B';
	pUSBDesc->usbStringDesc3.String[3]						= ' ';
	pUSBDesc->usbStringDesc3.String[4]						= 'C';
	pUSBDesc->usbStringDesc3.String[5]						= 'l';
	pUSBDesc->usbStringDesc3.String[6]						= 'i';
	pUSBDesc->usbStringDesc3.String[7]						= 'e';
	pUSBDesc->usbStringDesc3.String[8]						= 'n';
	pUSBDesc->usbStringDesc3.String[9]						= 't';
	pUSBDesc->usbStringDesc3.String[10] 					= ' ';
	pUSBDesc->usbStringDesc3.String[11] 					= 'S';
	pUSBDesc->usbStringDesc3.String[12] 					= 'a';
	pUSBDesc->usbStringDesc3.String[13] 					= 'm';
	pUSBDesc->usbStringDesc3.String[14] 					= 'p';
	pUSBDesc->usbStringDesc3.String[15] 					= 'l';
	pUSBDesc->usbStringDesc3.String[16] 					= 'e';
	pUSBDesc->usbStringDesc3.String[17] 					= '.';
}


HANDLE open_device()
{
	CHAR DevicePath[MAX_PATH];
	HANDLE hndFile;

	if( FALSE == GetDevicePath((LPGUID)&GUID_DEVINTERFACE_IOHUSBD,DevicePath))
	{
		printf("Unable to Find the Device Interface\n");
		return INVALID_HANDLE_VALUE;
	}

	hndFile = CreateFile(
					DevicePath,				/* Open the Device "file" using Dos device name */
					GENERIC_WRITE,
					FILE_SHARE_WRITE,
					NULL,
					OPEN_EXISTING,
					0,
					NULL);

	if (hndFile == INVALID_HANDLE_VALUE)	/* Was the device opened? */
	{
		printf("Unable to open the USB device.\n");
	}
	else
	{
		printf("OPENED SUCCESSFULLY.\n");
	}

	return hndFile;
}


HANDLE
open_file (
	PCHAR	filename
)
{
	CHAR DevicePath[MAX_PATH];
	size_t length;
	errno_t ret;
	HANDLE h;

	if( FALSE == GetDevicePath((LPGUID)&GUID_DEVINTERFACE_IOHUSBD,DevicePath))
	{
		printf("Unable to Find the Device Interface\n");
		return INVALID_HANDLE_VALUE;
	}

	ret = strcat_s(DevicePath, MAX_PATH, "\\");
	if(ret != 0)
	{
		printf("Failed to open handle - possibly long filename\n");
		return INVALID_HANDLE_VALUE;
	}

	ret = strcat_s(DevicePath, MAX_PATH, filename);
	if(ret != 0)
	{
		printf("Failed to open handle - possibly long filename\n");
		return INVALID_HANDLE_VALUE;
	}

	h = CreateFile(DevicePath,
				   GENERIC_WRITE | GENERIC_READ,
				   FILE_SHARE_WRITE | FILE_SHARE_READ,
				   NULL,
				   OPEN_EXISTING,
				   0,
				   NULL);

	if (h == INVALID_HANDLE_VALUE)
	{
		printf("Failed to open: Error = %d\n", (INT)GetLastError());
	}
	else
	{
		printf("Opened successfully.\n");
	}

	return h;
}


BOOL
SampleLoopback()
{
	USB_DESCRIPTOR_SAMPLE usbDescriptor;
	HANDLE hDevice, hRead, hWrite;
	ULONG nLength, nBytesRead, nBytesWrite;
	PVOID pBuffer;
	BOOL bResult;
	ULONG CurrentConfig;
	int iLoopCount;

	printf("\n");
	printf("USB Client controller sample program\n\n");
	pBuffer = malloc((BUFFER_SIZE/512+1)*512);
	if(pBuffer == NULL)
	{
		printf("No enough memory.\n");
		return	FALSE;
	}

	printf("Filling the USB Descriptors\n");
	usbDescriptorFill(&usbDescriptor);

	printf("\n");
	printf("Open device\n");
	hDevice = open_device();
	if(hDevice == INVALID_HANDLE_VALUE)
	{
		free(pBuffer);
		return	FALSE;
	}

	printf("\n");
	printf("IOCTL_SET_USB_DESCRIPTORS to give the device descriptors\n");
	bResult = DeviceIoControl(
				hDevice,					/* Handle to device. */
				IOCTL_SET_USB_DESCRIPTORS,	/* IO Control code for Data transfer */
				&usbDescriptor, 			/* Buffer to driver. Holds handle to EP/ */
				sizeof(usbDescriptor),		/* Length of buffer in bytes. */
				NULL,						/* Buffer from driver. Not used. */
				0,							/* Length of buffer in bytes. */
				&nLength,					/* Bytes placed in outbuf.	Should be 0. */
				NULL						/* NULL means wait till I/O completes. */
				);
	if(bResult == 1)
	{
		printf("SUCCESS\n");
	}
	else
	{
		printf("FAILED\n");
		CloseHandle(hDevice);
		free(pBuffer);
		return	FALSE;
	}

	printf("\n");
	printf("IOCTL_GET_CONFIG_NUMBER to wait for setting a device configuration\n");
	CurrentConfig = 0;
	while(CurrentConfig==0)
	{
		bResult = DeviceIoControl(
			hDevice,						/* Handle to device. */
			IOCTL_USB_GET_CONFIG_NUMBER,	/* IO Control code for Data transfer */
			NULL,							/* Buffer to driver. Holds handle to EP/ */
			0				,				/* Length of buffer in bytes. */
			&CurrentConfig, 				/* Buffer from driver. Not used. */
			sizeof(CurrentConfig),			/* Length of buffer in bytes. */
			&nLength,						/* Bytes placed in outbuf.	Should be 0. */
			NULL							/* NULL means wait till I/O completes. */
			);
		if(bResult == 1)
		{
			if(CurrentConfig == 1)
			{
				printf("SUCCESS\n");
				break;
			}
		}
		else
		{
			printf("FAILED\n");
			CloseHandle(hDevice);
			free(pBuffer);
			return	FALSE;
		}

		Sleep(1000);
	}

	printf("\n");
	printf("Open pipe for OUT transfer\n");
	hRead = open_file("2"); 	/* Number of EP1 OUT */
	if(hRead == INVALID_HANDLE_VALUE)
	{
		printf("FAILED\n");
		CloseHandle(hDevice);
		free(pBuffer);
		return	FALSE;
	}

	printf("\n");
	printf("Open pipe for IN transfer\n");
	hWrite = open_file("3");	/* Number of EP1 IN */
	if(hWrite == INVALID_HANDLE_VALUE)
	{
		printf("FAILED\n");
		CloseHandle(hRead);
		CloseHandle(hDevice);
		free(pBuffer);
		return	FALSE;
	}

	iLoopCount = 0;
	while(1)
	{

		printf("\n");
		printf("Request USB OUT Transtfer\n");
		bResult = ReadFile( hRead,
							pBuffer,
							BUFFER_SIZE,
							&nBytesRead,
							NULL);
		if(bResult)
		{
			printf("ReadFile : request %06.6d bytes -- %06.6d bytes read\n",
			(INT)BUFFER_SIZE, (INT)nBytesRead);
		}
		else
		{
			printf("ReadFile failed with error code %d\n", (INT)GetLastError());
			CloseHandle(hWrite);
			CloseHandle(hRead);
			CloseHandle(hDevice);
			free(pBuffer);
			return	FALSE;
		}

		printf("\n");
		printf("Request USB IN Transtfer\n");
		bResult = WriteFile(hWrite,
							pBuffer,
							nBytesRead,
							&nBytesWrite,
							NULL);
		if(bResult)
		{
			printf("WriteFile : request %06.6d bytes -- %06.6d bytes wrote\n",
			(INT)nBytesRead, (INT)nBytesWrite);
		}
		else
		{
			printf("WriteFile failed with error code %d\n", (INT)GetLastError());
			CloseHandle(hWrite);
			CloseHandle(hRead);
			CloseHandle(hDevice);
			free(pBuffer);
			return	FALSE;
		}

		iLoopCount++;
		printf("loop count = %d\n", iLoopCount);
		printf("************************************************************\n\n");
	}
}


int __cdecl
main(
	__in int argc,
	__in_ecount(argc) char *argv[]
	)
{
	SampleLoopback();
	return 0;
}
